"""通用的 report data 构造器，支持 校、区、市、省、全国级别的通用报告数据构造"""
from dataclasses import dataclass, field
from typing import Optional, ClassVar, Any, Set

import pandas as pd
import pendulum

from hawa.base.db import DbUtil, RedisUtil
from hawa.common.query import DataQuery
from hawa.common.utils import GradeData, CaseData, Measurement
from hawa.config import project


class MetaCommomData(type):
    def __new__(cls, name, bases, attrs):
        attrs['db'] = DbUtil()
        attrs['redis'] = RedisUtil()
        attrs['query'] = DataQuery()
        return super().__new__(cls, name, bases, attrs)


@dataclass
class CommonData(metaclass=MetaCommomData):
    # 构造单位
    meta_unit_type: Optional[str] = ''  # school/district/city/province/country
    meta_unit_id: Optional[int] = None
    meta_unit: Optional[Any] = None

    # 时间目标
    target_year: Optional[int] = None
    last_year_num: Optional[int] = None
    is_load_last: bool = True  # 仅计算往年数据时 为 False

    # 卷子
    test_type: str = ''
    test_types: list = field(default_factory=list)
    code_word_list: Set[str] = field(default=set)  # 卷子使用指标的词表，详见继承

    # meta class tool
    db: ClassVar[DbUtil] = None
    redis: ClassVar[RedisUtil] = None
    query: ClassVar[DataQuery] = None

    # 原始数据

    school_ids: list[int] = field(default_factory=list)
    schools: pd.DataFrame = pd.DataFrame()

    cases: pd.DataFrame = pd.DataFrame()
    case_ids: list[int] = field(default_factory=list)

    answers: pd.DataFrame = pd.DataFrame()

    students: pd.DataFrame = pd.DataFrame()
    student_ids: list[int] = field(default_factory=list)
    student_count: Optional[int] = None

    item_ids: Optional[set[int]] = None
    items: Optional[pd.DataFrame] = None

    # 辅助工具
    grade_util: Optional[GradeData] = None
    case_util: Optional[CaseData] = None
    measurement = Measurement()

    # 计算数据
    final_answers: pd.DataFrame = pd.DataFrame()
    final_scores: pd.DataFrame = pd.DataFrame()

    # 去年全国数据
    last_year = None
    last_year_code_scores: Optional[pd.DataFrame] = pd.DataFrame()

    def __post_init__(self):
        # 初始化数据
        init_functions = [i for i in dir(self) if i.startswith('_to_init_')]
        for func in init_functions:
            getattr(self, func)()

        # 构建辅助工具
        self._to_build_helper()

        # 计算数据
        count_functions = [i for i in dir(self) if i.startswith('_to_count_')]
        for func in count_functions:
            getattr(self, func)()

    def _to_init_a_meta_unit(self):
        self.meta_unit = self.query.query_unit(self.meta_unit_type, self.meta_unit_id)

    def _to_init_b_time(self):
        if not self.target_year:
            self.target_year = pendulum.now().year
        self.last_year_num = self.target_year - 1
        project.logger.info(f'target_year: {self.target_year}')

    def _to_init_c_schools(self):
        if self.school_ids:
            self.schools = self.query.query_schools_by_ids(self.school_ids)
        else:
            if self.meta_unit_type == 'country':
                self.schools = self.query.query_schools_all()
            else:
                self.schools = self.query.query_schools_by_startwith(self.meta_unit_id)
            self.school_ids = self.schools['id'].tolist()
        project.logger.debug(f'schools: {len(self.schools)}')

    def _to_init_d_cases(self):
        start_stamp = pendulum.datetime(self.target_year, 1, 1)
        end_stamp = pendulum.datetime(self.target_year + 1, 1, 1)
        start_stamp_str = start_stamp.format(project.format)
        end_stamp_str = end_stamp.format(project.format)

        papers = self.query.query_papers(test_types=self.test_types, test_type=self.test_type)
        paper_ids = papers['id'].tolist()

        self.cases = self.query.query_cases(
            school_ids=self.school_ids,
            paper_ids=paper_ids,
            valid_to_start=start_stamp_str,
            valid_to_end=end_stamp_str,
        )
        if self.cases.empty:
            raise Exception(f'no cases:{self.meta_unit} {self.school_ids}')
        self.case_ids = self.cases['id'].tolist()
        project.logger.debug(f'cases: {len(self.cases)}')

    def _to_init_e_answers(self):
        self.answers = self.query.query_answers(case_ids=self.case_ids)
        project.logger.debug(f'answers: {len(self.answers)}')

    def _to_init_f_students(self):
        self.student_ids = set(self.answers['student_id'].tolist())
        student_id_list = list(self.student_ids)
        self.students = self.query.query_students(student_id_list)
        self.student_count = len(self.student_ids)
        project.logger.debug(f'students: {self.student_count}')

    def _to_init_g_items(self):
        self.item_ids = set(self.answers['item_id'].drop_duplicates())
        self.items = self.query.query_items(self.item_ids)
        project.logger.debug(f'items: {len(self.items)}')

    def _to_build_helper(self):
        self.grade = GradeData(case_ids=self.case_ids)
        self.case = CaseData(cases=self.cases)

    def _to_count_a_final_answers(self):
        items = {}
        item_codes = self.query.query_item_codes(self.item_ids)
        project.logger.debug(f"{self.code_word_list=}")
        for item_id, codes in item_codes.groupby('item_id'):
            items[item_id] = [i for i in codes.to_dict(orient='records')
                              if i['category'] in self.code_word_list]

        data = pd.merge(
            self.answers, self.students.loc[:, ['id', 'gender', 'first_name', 'last_name']],
            left_on='student_id', right_on='id'
        )
        project.logger.debug(f"ans merge students {len(data)}")
        # inner 时，final_answers 和 answers 数目不等：final_answers 过滤掉了 没有 code_word_list（维度领域或其他）的题目
        # outer 时，数目相等，不过滤任何题目
        data = pd.merge(data, item_codes, left_on='item_id', right_on='item_id', how='inner')

        data['cls'] = data['id_y'].apply(lambda x: int(str(x)[13:15]))
        data['grade'] = data['case_id'].apply(lambda x: int(str(x)[-2:]))
        data['username'] = data.apply(lambda x: f"{x.last_name}{x.first_name}", axis=1)
        for code_word in self.code_word_list:
            data[code_word] = data.item_id.apply(self._count_field, args=(code_word, items))
        self.final_answers = data.drop_duplicates(subset=['case_id', 'student_id', 'item_id'])
        project.logger.debug(f'final_answers: {len(self.final_answers)}')

    def _to_count_b_final_scores(self):
        self.final_scores = self.count_final_score(answers=self.final_answers)
        project.logger.debug(f'final_scores: {len(self.final_scores)}')

    def _count_field(self, item_id: int, code: str, items: dict):
        return [i for i in items[item_id] if i['category'] == code][0]['name']

    @staticmethod
    def count_level(score, mode: str = 'f'):
        assert mode in ('f', 'r'), 'only support feedback or report'
        if score >= 90:
            a = 'A'
        elif score >= 80:
            a = 'B'
        elif score >= 60:
            a = 'C'
        else:
            a = 'D'
        key = "RANK_LABEL" if mode == 'r' else 'FEEDBACK_LEVEL'
        return project.ranks[key][a]

    def count_final_score(self, answers: pd.DataFrame):
        res = answers.groupby(by=['grade', 'gender', 'student_id']).score.mean().reset_index()
        res = res.assign(score=res.score * 100)
        res['level'] = res.score.apply(lambda x: self.count_level(x))
        return res



from loguru import logger

from hawa.common.data import CommonData
from test.mock import prepare_test

prepare_test()


def test_common_init():
    rows = [
        {"meta_unit_type": "school", "meta_unit_id": 3707030003, "target_year": 2021},
        # {"meta_unit_type": "district", "meta_unit_id": 370703, "target_year": 2021},
        # {"meta_unit_type": "city", "meta_unit_id": 331100, "target_year": 2021},
        # {"meta_unit_type": "province", "meta_unit_id": 410000, "target_year": 2021},
    ]
    for row in rows:
        logger.info(row)
        # common data without default  test_types and code_word_list
        CommonData(
            **row, test_types=['publicWelfare', 'ZjpublicWelfare'],
            code_word_list={'dimension', 'field'}
        )



import inspect

from hawa.config import project


def build_set_project_params(p):
    """p 为 config.Settings"""
    f = inspect.signature(set_project)
    params = f.parameters
    func_params = {k: getattr(p, k.upper()) for k in params if k != 'self'}
    return func_params


def set_project(
        db_host: str = '',
        db_port: int = 3306,
        db_user: str = '',
        db_pswd: str = '',
        db_name: str = '',
        redis_host: str = '',
        redis_db: int = 0,
        mongo_host: str = '',
        mongo_port: int = 27017,
        mongo_user: str = '',
        mongo_pswd: str = '',
        mongo_db: str = '',
        mongo_auth_db: str = '',
):
    for k, v in vars().items():
        upper_k = k.upper()
        if upper_k in dir(project):
            setattr(project, upper_k, v)
    project.COMPLETED = True



from hawa.paper.mht import MhtData
from test.mock import prepare_test

prepare_test()


def test_mht_init():
    rows = [
        {"meta_unit_type": "school", "meta_unit_id": 4107110001,
         "target_year": 2022, "test_type": "mht"},
    ]
    for row in rows:
        md = MhtData(**row)
        assert len(md.scale_student_score) == 3
        assert len(md.sub_scale_score) == 3
        assert len(md.grade_scale_student_score) == 3
        assert len(md.grade_special_students) == 3




import platform

from mongo_log.models import LogRecord
from pydantic import BaseSettings


def is_debug():
    system = platform.system()
    dev_system = ('Windows', 'Darwin')
    return True if system in dev_system else False


class Settings(BaseSettings):
    # mode
    DEBUG = is_debug()
    COMPLETED = False  # 配置是否加载完成

    PROJECT = 'HawaData'
    client_id = 6

    # db
    DB_MODE = 'mysql'
    DB_HOST = '' if not DEBUG else 'localhost'
    DB_PORT = 3306
    DB_NAME = '' if not DEBUG else 'yass'
    DB_USER = '' if not DEBUG else 'test'
    DB_PSWD = '' if not DEBUG else 'test'

    # redis
    REDIS_HOST = '' if not DEBUG else 'localhost'
    REDIS_DB = 5
    REDIS_PREFIX = 'report:'

    # mongo
    MONGO_HOST = '' if not DEBUG else 'localhost'
    MONGO_PORT = 3717 if not DEBUG else 27017
    MONGO_USER = '' if not DEBUG else 'test'
    MONGO_PSWD = '' if not DEBUG else 'test'
    MONGO_DB = 'record'
    MONGO_AUTH_DB = 'admin'

    # time
    utc = 'UTC'
    timezone = 'Asia/Shanghai'
    format = 'YYYY-MM-DD HH:mm:ss'

    # case
    time_distribution_minutes = 30
    min_join_number = 10
    min_alone_numger = 5

    gender_map = {
        'total': '总体', 'M': '男生', "F": "女生"
    }

    # rank
    ranks = {
        'RANK_LABEL': {'A': '优秀', 'B': '良好', 'C': '达标', 'D': '待达标'},
        'FEEDBACK_LEVEL': {'A': '优秀', 'B': '良好', 'C': '中等', 'D': '待提高'}
    }
    reverse_ranks = {
        'RANK_LABEL': {k: v for v, k in ranks['RANK_LABEL'].items()},
        'FEEDBACK_LEVEL': {k: v for v, k in ranks['FEEDBACK_LEVEL'].items()},
    }

    # grades
    grade_map = {
        1: '一', 2: '二', 3: '三', 4: '四', 5: '五', 6: '六', 7: '初中一',
        8: '初中二', 9: '初中三', 10: '高中一', 11: '高中二',
        12: '高中三'
    }
    grade_simple = {**dict(zip(range(1, 11), '一二三四五六七八九十')), **{11: '十一', 12: '十二'}}

    # number
    number_map = dict(zip(range(1, 11), '一二三四五六七八九十'))

    category_map = {
        'total': "学生", 'M': '男生', "F": "女生"
    }

    # other
    municipality = {11, 12, 31, 50}

    class Config:
        env_prefix = 'No_Env_'  # 不使用环境变量

    @property
    def grade_mapping(self):
        mappings = dict(zip(range(0, 13), '无 一 二 三 四 五 六 七 八 九 十 十一 十二'.split(' ')))
        return mappings

    @property
    def logger(self):
        return LogRecord(project=self.PROJECT)


project = Settings()



import random
from collections import defaultdict
from dataclasses import dataclass

import pandas as pd
from munch import Munch

from hawa.config import project
from hawa.paper.health import HealthReportData
from hawa.paper.mht import MhtWebData


@dataclass
class SchoolMixin:
    """为了在 __mro__ 中有更高的优先级， mixin 在继承时，应该放在最前"""
    meta_unit_type: str = 'school'


@dataclass
class SchoolHealthReportData(SchoolMixin, HealthReportData):
    grade_gender_distribution = None
    grade_score = None  # 年级最高、最低、平均分
    grade_rank_dis = None  # 年级、性别、水平学生分布占比
    grade_reverse_rank_dis = None  # 占比由高到低的年级分布水平
    grade_code_score = None  # 健康素养水平 所需数据
    compare_grade_year_school = None  # 比较学校与去年维度领域
    compare_all_total = None  # 比照去年数据和学校数据的全年级平均数

    async def _to_count_k_grade_reverse_rank_dis(self):
        records = {}
        rank_dis = self.grade_rank_dis
        for g in self.grade.grades:
            here = rank_dis[g].total
            base = [(v, k) for k, v in here.items()]
            records[g] = [b for b in sorted(base, key=lambda x: x[0], reverse=True)]
        self.grade_reverse_rank_dis = records

    async def _to_count_l_grade_code_score(self):
        """健康素养水平 所需数据"""
        records = {}
        scores = self.code_scores
        for grade in self.grade.grades:
            base: pd.DataFrame = scores.loc[scores.grade == grade, :]
            base_dimensions = base.loc[base.category == 'dimension', ['code', 'total']]
            base_fields = base.loc[base.category == 'field', ['code', 'total']]

            dimensions = self._count_df_reverse(first_col='code', second_col='total', data=base_dimensions)
            fields = self._count_df_reverse(first_col='code', second_col='total', data=base_fields)

            a = base.loc[base.category == 'field', ['code', 'F']].mean().mean()
            b = base.loc[base.category == 'field', ['code', 'M']].mean().mean()
            records[grade] = {
                'dimension': dimensions, 'field': fields,
                'cond': self.count_cond(a, b)
            }
        self.grade_code_score = records

    async def _to_count_n_compare_grade_year_school(self):
        if not self.is_load_last:
            return
        res = defaultdict(dict)
        cols = ['total', 'M', 'F']
        for grade in self.grade.grades:
            for col in cols:
                res[grade][col] = self._count_dim_field_diff(grade=grade, key=col)
        self.compare_grade_year_school = res

    async def _to_count_o_compare_all_total(self):
        if not self.is_load_last:
            return
        data = []
        for grade in self.grade.grades:
            sch = self.grade_score[grade].avg
            year = self.last_year[str(grade)]['score']['total']
            data.append(sch - year)
        self.compare_all_total = self.count_cond(a=sum(data) / len(data), b=0)

    # property 数据
    @property
    def gender_count(self) -> Munch:
        r = self.students.gender.value_counts()
        return Munch({'M': 0, 'F': 0} | r.to_dict())

    # 工具函数

    def _count_df_reverse(self, first_col: str, second_col: str, data: pd.DataFrame):
        base = data.to_dict(orient='record')
        middle = {d[first_col]: d[second_col] / 100 for d in base}
        reverse_middle = {v: k for k, v in middle.items()}
        res = [(self._retain_prec(k), reverse_middle[k]) for k in sorted(reverse_middle.keys(), reverse=True)]
        return res

    def count_cond(self, a: float, b: float):
        if a == b:
            conditions = '等于'.split()
        elif a - b >= 5:
            conditions = '明显高于'.split()
        elif abs(a - b) < 5:
            conditions = '差异不明显于'.split()
        elif a - b <= -5:
            conditions = '明显低于'.split()
        else:
            raise
        return random.choice(conditions)

    def _count_dim_field_diff(self, grade: int, key: str):
        """计算学校与全国对比的维度、领域高低
        :param key: 比较的项：total/M/F
        """
        data = self._build_one_data(grade=grade)
        key_col = f'{key}_y_s'
        data[key_col] = data[f"{key}_sch"] - data[f"{key}_year"]
        cols = ['category_sch', 'code', key_col, f'{key}_year', f'{key}_sch']
        compare = data.sort_values(['category_sch', key_col]).loc[:, cols]
        dim_compare = compare.loc[compare.category_sch == 'dimension', :]
        field_compare = compare.loc[compare.category_sch == 'field', :]
        dim_low, dim_high = dim_compare.head(n=1), dim_compare.tail(n=1)
        field_low, field_high = field_compare.head(n=1), field_compare.tail(n=1)
        return {
            'dim_low': dim_low.code.values[0] if dim_low[key_col].values[0] < 0 else '',
            'field_low': field_low.code.values[0] if field_low[key_col].values[0] < 0 else '',
            'dim_high': dim_high.code.values[0] if dim_high[key_col].values[0] > 0 else '',
            'field_high': field_high.code.values[0] if field_high[key_col].values[0] > 0 else '',
        }

    def _build_one_data(self, grade: int):
        """抽象某个构建数据的逻辑"""
        year_scores = self.last_year_code_scores
        sch_scores = self.code_scores
        year_code_score = year_scores.loc[year_scores.grade == grade, :]
        school_code_score = sch_scores.loc[sch_scores.grade == grade, :]
        data = pd.merge(
            year_code_score, school_code_score, left_on='code', right_on='code',
            suffixes=['_year', '_sch']
        )
        return data

    def build_code_gender_data(self, category: str, g: int):
        data = self._build_one_data(grade=g)
        use_cols = [f'{category}_year', f"{category}_sch", 'code', 'category_sch']
        data = data.loc[:, use_cols].sort_values(['category_sch', 'code'])
        del data['category_sch']
        return data

    def compare_gender_text(self, grade: int):
        grade_rank_dis_m = self.grade_rank_dis[grade].M
        grade_rank_dis_f = self.grade_rank_dis[grade].F
        level_m = round(sum([grade_rank_dis_m['优秀'], grade_rank_dis_m['良好']]), 1)
        level_f = round(sum([grade_rank_dis_f['优秀'], grade_rank_dis_f['良好']]), 1)
        if level_m - level_f >= 5:
            return '男生明显大于女生'
        elif level_m - level_f <= -5:
            return '男生明显小于女生'
        else:
            return '男生与女生不存在明显差异'

    def low_high_code_text(self, grade: int, gender: str):
        """相对最高、相对最低文本
        :param grade: total/M/F
        :param gender: total/M/F
        """
        if not self.is_load_last:
            return
        data = self.build_code_gender_data(category=gender, g=grade)
        data['diff'] = data[f"{gender}_sch"] - data[f"{gender}_year"]
        bigger = data.loc[data['diff'] >= 5, :].code.to_list()
        smaller = data.loc[data['diff'] <= -5, :].code.to_list()
        res = ''
        if bigger:
            local_codes = self._build_codes(bigger)
            res += f"{local_codes}{self.meta_unit.short_name}{project.category_map[gender]}分数" \
                   f"明显高于全国{project.grade_simple[grade]}年级{project.category_map[gender]}平均分数，"
        if smaller:
            local_codes = self._build_codes(smaller)
            res += f"{local_codes}{self.meta_unit.short_name}{project.category_map[gender]}分数" \
                   f"明显低于全国{project.grade_simple[grade]}年级{project.category_map[gender]}平均分数，"
        if not bigger and not smaller:
            res += '所有维度和领域都没有明显差异。'
        else:
            res += '其余维度和领域没有明显差异。'
        return res

    def _build_codes(self, codes: list[str]):
        if len(codes) == 1:
            return f'“{codes[0]}”'
        elif len(codes) > 1:
            return '、'.join([f'“{i}”' for i in codes[:-1]]) + f'和“{codes[-1]}”'
        else:
            return ''

    def compare_grade_gen_text(self):
        """生成整体分析标题及一句分析文本。"""
        data = {}
        for grade in self.grade.grades:
            sch = self.grade_score[grade].avg
            year = self.last_year[str(grade)]['score']['total']
            data[grade] = sch - year
        total_dif = sum(data.values()) / len(data)
        diff = []
        if total_dif > 0:
            for k, v in data.items():
                if v < 0:
                    diff.append(k)
        elif total_dif < 0:
            for k, v in data.items():
                if v > 0:
                    diff.append(k)
        else:
            pass
        cond = self.compare_all_total
        if diff:
            grade_text = '除' + '、'.join([project.grade_simple[g] for g in diff]) + '年级外，'
        else:
            grade_text = ''
        title_text = f"{grade_text}各年级生命与健康素养{cond}全国平均水平"
        describe_text = f"{grade_text}{self.meta_unit.short_name}各年级的健康素养水平{cond}{self.last_year_num}年的全国平均水平"
        return title_text, describe_text

    def describe_grade_text(self, category: str):
        """描述全年级对比情况
        :param category: total/gender  total：总体比全国 / gender：男生比女生。
        """
        bigger, smaller = [], []
        for grade in self.grade.grades:
            if category == 'total':
                first = self.grade_score[grade].avg
                second = self.last_year[str(grade)]['score']['total']
            else:
                first = self.summary_scores[grade]['M']
                second = self.summary_scores[grade]['F']
            if first - second >= 5:
                bigger.append(f"{project.grade_simple[grade]}年级")
            elif first - second <= -5:
                smaller.append(f"{project.grade_simple[grade]}年级")
        res = ''
        if bigger:
            local_codes = self._build_codes(bigger)
            res += f"{self.meta_unit.short_name}{local_codes}分数"
            match category:
                case 'total':
                    res += '明显高于全国平均分数，'
                case 'gender':
                    res += '男生明显高于女生，'

        if smaller:
            local_codes = self._build_codes(smaller)
            res += f"{self.meta_unit.short_name}{local_codes}分数"
            match category:
                case 'total':
                    res += "明显低于全国平均分数，"
                case 'gender':
                    res += "男生明显低于女生，"
        if (bigger or smaller) and len(bigger) + len(smaller) < len(self.grade.grades):
            res += '其他年级没有明显差异。'
        if not bigger and not smaller:
            res = f'{self.meta_unit.short_name}分数对比无明显差异。'
        return res

    def replace_hai(self, text: str, condition: list):
        if condition:
            return text
        else:
            return text.replace('还', '')


@dataclass
class SchoolMhtWebData(SchoolMixin, MhtWebData):
    meta_unit_type: str = 'school'
    pass



import pendulum

from hawa.common.query import DataQuery
from hawa.config import project
from test.mock import prepare_test

prepare_test()

dq = DataQuery()


def test_query_unit():
    rows = [
        ['school', 1101019055], ["district", 110101], ["city", 331100],
        ["province", 110000], ["country", 0]]
    for r in rows:
        res = dq.query_unit(*r)
        assert res.id == r[1]


def test_query_schools():
    rows = [[1101019055], [1101019055, 1101019056]]
    for r in rows:
        res = dq.query_schools_by_ids(r)
        assert len(res) >= 1

    startwith = 110101
    res = dq.query_schools_by_startwith(startwith)
    assert len(res) >= 1

    res = dq.query_schools_all()
    assert len(res) >= 1


def test_query_papers():
    conditions = [
        {'test_type': 'mht'}, {'test_type': 'psychological'},
        {'test_types': ['publicWelfare', 'ZjpublicWelfare']}
    ]
    for c in conditions:
        res = dq.query_papers(**c)
        assert len(res) >= 1


def test_query_cases():
    target_year = 2021
    start_stamp = pendulum.datetime(target_year, 1, 1)
    end_stamp = pendulum.datetime(target_year + 1, 1, 1)
    start_stamp_str = start_stamp.format(project.format)
    end_stamp_str = end_stamp.format(project.format)

    papers = dq.query_papers(test_types=['publicWelfare', 'ZjpublicWelfare'])
    paper_ids = [i[1]['id'] for i in papers.iterrows()]

    school_ids = [3707850002, 3707030003]
    res = dq.query_cases(school_ids, paper_ids, start_stamp_str, end_stamp_str)
    assert len(res) >= 1
    res = dq.query_cases(school_ids[:1], paper_ids, start_stamp_str, end_stamp_str)
    assert len(res) >= 1


def test_query_answers():
    conditions = [
        [2209250708],
        [2209250708, 2209250803]
    ]
    for c in conditions:
        res = dq.query_answers(c)
        assert len(res) >= 1


def test_query_students():
    case_ids = 2210240104
    answers = dq.query_answers([case_ids])
    res = dq.query_students(answers.student_id.tolist())
    assert len(res) >= 1


def test_query_items():
    res = dq.query_items([1, 401])
    assert len(res) >= 1


def test_query_item_codes():
    res = dq.query_item_codes([1, 401])
    assert len(res) >= 1



from loguru import logger

from hawa.data.school import SchoolHealthReportData, SchoolMhtWebData
from test.mock import prepare_test, validate_data_for_web

prepare_test()


def test_health_report_run():
    rows = [
        {"meta_unit_id": 3707030003, "target_year": 2021},
    ]
    for row in rows:
        logger.info(row)
        SchoolHealthReportData(**row)


def test_mht_web_run():
    rows = [
        {"meta_unit_id": 4107110001, "target_year": 2022},
    ]
    for row in rows:
        md = SchoolMhtWebData(**row)
        assert len(md.scale_student_score) == 3
        assert len(md.sub_scale_score) == 3
        assert len(md.grade_scale_student_score) == 3
        assert len(md.grade_special_students) == 3

        data = [
            md.scale_student_score, md.sub_scale_score,
            md.grade_scale_student_score, md.grade_sub_scale_score,
            md.grade_special_students
        ]
        for d in data:
            validate_data_for_web(d)

        assert len(md.scale_student_score['x_axis']) == 100
        assert len(md.scale_student_score['y_axis']) == 100
        print(md.sub_scale_score)

        print(md.grade_sub_scale_score)



"""健康测评数据"""
import itertools
import json
from collections import defaultdict
from dataclasses import dataclass, field
from typing import Union, Set

import pandas as pd
from munch import Munch

from hawa.common.data import CommonData
from hawa.config import project


@dataclass
class HealthData(CommonData):
    """健康测评数据，不应直接使用，应向下继承 city/district/school 等"""
    test_types: list[str] = field(default_factory=lambda: ['publicWelfare', 'ZjpublicWelfare'])
    code_word_list: Set[str] = field(default_factory=lambda: {'dimension', 'field'})


@dataclass
class HealthReportData(HealthData):
    # 计算数据
    code_scores: pd.DataFrame = pd.DataFrame()
    summary_scores: dict = field(default_factory=dict)
    grade_good_bad = None  # 优势 优先关注点

    def _to_count_a0_last_year_data(self):
        if not self.is_load_last:
            return
        key = f'{project.REDIS_PREFIX}{self.last_year_num}:data'
        if not self.redis.conn.exists(key):
            raise ValueError(f'last year data not exists: {key}')
        self.last_year = json.loads(self.redis.conn.get(key))
        res = []
        for grade in self.grade.grades:
            year_code_score = pd.DataFrame(self.last_year[str(grade)]['code']).T
            year_code_score['i'] = [int(grade) * 10 + i for i in range(1, 11)]
            year_code_score.set_index('i', inplace=True)
            res.append(year_code_score)
        self.last_year_code_scores = pd.concat(res)

    def _to_count_c_code_scores(self):
        records = []
        for grade in self.grade.grades:
            for code in self.measurement.dimensions:
                record = self._count_grade_code_score(grade=grade, code=code, category='dimension')
                records.append(record)
            for code in self.measurement.fields:
                record = self._count_grade_code_score(grade=grade, code=code, category='field')
                records.append(record)
        self.code_scores = pd.DataFrame.from_records(records)

    def _to_count_d_summary_scores(self):
        records = defaultdict(dict)
        for grade, group in self.code_scores.groupby('grade'):
            for gender in project.gender_map.keys():
                records[grade][gender] = round(self._count_school_score(group, key=gender), 1)
        self.summary_scores = records

    def _to_count_e_grade_good_bad(self):
        records = defaultdict(dict)
        for g, group in self.code_scores.groupby('grade'):
            for gender in project.gender_map.keys():
                records[g][gender] = {
                    'dimension': {'good': [], 'bad': []},
                    'field': {'good': [], 'bad': []}
                }
            for _, row in group.iterrows():
                for gender in project.gender_map.keys():
                    if row[gender] >= 80:
                        if row['category'] == 'dimension':
                            records[g][gender]['dimension']['good'].append(row['code'])
                        if row['category'] == 'field':
                            records[g][gender]['field']['good'].append(row['code'])
                    elif row[gender] < 60:
                        if row['category'] == 'dimension':
                            records[g][gender]['dimension']['bad'].append(row['code'])
                        if row['category'] == 'field':
                            records[g][gender]['field']['bad'].append(row['code'])
                    else:
                        pass
                else:
                    for code in itertools.chain(self.measurement.dimensions, self.measurement.fields):
                        for ccc in ['dimension', 'field']:
                            if code in records[g]['total'][ccc]['bad']:
                                for ggg in 'MF':
                                    try:
                                        records[g][ggg][ccc]['bad'].remove(code)
                                    except ValueError:
                                        pass

        self.grade_good_bad = records

    def _to_count_f_case_gender_counts(self):
        """单年级，各班级性别数据；多年级，各年级性别数据。"""
        ans = self.final_answers.drop_duplicates(subset=['student_id'])
        data = ans.loc[:, ['grade', 'cls', 'gender', 'student_id']]
        # 年级 班级 男生数 女生数 总人数
        # 班级 男生数 女生数 总人数
        records = []
        for grade, grade_group in data.groupby(by=['grade']):
            for cls, cls_group in grade_group.groupby('cls'):
                grade_cls = Munch(grade=grade, cls=cls)
                cls_total = Munch(total=len(cls_group))
                cls_gender = cls_group.gender.value_counts().to_dict()
                records.append(grade_cls | cls_total | cls_gender)
            grade_total = Munch(total=len(grade_group))
            grade_gender = grade_group.gender.value_counts().to_dict()
            records.append({'grade': grade, 'cls': 0, 'F': 0, 'M': 0} | grade_total | grade_gender)
        records = pd.DataFrame.from_records(records).fillna(0)
        records['M'] = records.M.astype('int')
        records['F'] = records.F.astype('int')
        if len(self.case_ids) == 1:
            records['cls'] = records.apply(self._count_cls, axis=1)
        else:
            records['grade'] = records.grade.apply(lambda x: f"{project.grade_simple[x]}年级")
            records = records.loc[records.cls == 0, :]
        self.case_gender_counts = records.to_dict(orient='records')

    def _to_count_g_cronbach_alpha(self):
        import pingouin as pg
        cols = ['student_id', 'item_id', 'score']
        res = []
        for grade, group in self.final_answers.groupby('grade'):
            base = group.loc[:, cols]
            data = pd.pivot_table(base, index='item_id', columns='student_id', values='score')
            c: tuple = pg.cronbach_alpha(data)
            res.append(c[0])
        self.cronbach_alpha = [round(i, 3) for i in res]

    def _to_count_h_grade_gender_distribution(self):
        """年级性别分布"""
        data = self.case_gender_counts
        records = {}
        if len(self.case_ids) == 1:
            for row in data:
                if '年级' in row['cls']:
                    records[row['cls']] = Munch(row)
        for row in data:
            records[row['grade']] = Munch(row)
        self.grade_gender_distribution = records

    def _to_count_i_grade_score(self):
        """年级最高、最低、平均分"""
        records = Munch()
        for grade in self.grade.grades:
            ans = self.final_answers.loc[self.final_answers.grade == grade, :]
            score = ans.groupby('student_id').score
            record = Munch(
                avg=self._retain_prec(score.mean().mean()),
                min=self._retain_prec(score.mean().min()),
                max=self._retain_prec(score.mean().max()))
            records[grade] = record
        self.grade_score = records

    def _to_count_j_grade_rank_dis(self):
        records = {}
        base = dict(
            zip(project.ranks['FEEDBACK_LEVEL'].values(),
                [0] * len(project.ranks['FEEDBACK_LEVEL'])))
        for grade, group in self.final_scores.groupby('grade'):
            count = base | group.level.value_counts().to_dict()
            records[grade] = Munch()
            for gender, g in group.groupby('gender'):
                gender_count = base | g.level.value_counts().to_dict()
                records[grade][gender] = \
                    {k: self._retain_prec(v / sum(gender_count.values()))
                     for k, v in gender_count.items()}
            records[grade].total = {k: self._retain_prec(v / sum(count.values())) for k, v in count.items()}
        self.grade_rank_dis = records

    def cache_year_data(self, year: int):
        """缓存年数据"""
        project.logger.info(f"缓存年数据 {year=}")
        key = f'{project.REDIS_PREFIX}{self.last_year_num}:data'

        res = defaultdict(dict)
        tool = self.__class__(meta_unit_type='country', meta_unit_id=0, target_year=year, is_load_last=False)
        project.logger.debug('after init data')

        for grade, value in zip(tool.grade.grades, tool.grade_gender_distribution.values()):
            res[grade]['people'] = value
        for grade, group in tool.code_scores.groupby(['grade']):
            res[grade]['code'] = {}
            for r in group.to_dict(orient='record'):
                res[grade]['code'][r['code']] = r
        for grade, value in tool.grade_rank_dis.items():
            res[grade]['rank'] = value
        score_dict = defaultdict(dict)
        # 计算年级、性别 平均分
        for grade, first_value in res.items():
            for second_key, second_value in first_value.items():
                if second_key == 'code':
                    total_score = self._count_year_score('total', second_value)
                    f_score = self._count_year_score('F', second_value)
                    m_score = self._count_year_score('M', second_value)
                    score_dict[grade] = {
                        'total': total_score, 'F': f_score, 'M': m_score
                    }
        for grade, value in score_dict.items():
            res[grade]['score'] = value
        # 计算总人数
        res['total'] = defaultdict(dict)
        res['total']['total'] = sum([self._get_value(v, 'total') for v in res.values()])
        res['total']['M'] = sum([self._get_value(v, 'M') for v in res.values()])
        res['total']['F'] = sum([self._get_value(v, 'F') for v in res.values()])
        self.redis.conn.set(key, json.dumps(res))

    def _count_grade_code_score(self, code: str, grade: int, category: str):
        """计算指定年级、指定维度、领域的分数"""
        ans = self.final_answers
        local_ans = ans.loc[(ans.grade == grade) & (ans[category] == code), :]
        student_score = local_ans.groupby('student_id').score.mean().mean() * 100
        gender_score = local_ans.groupby(['gender', 'student_id']) \
                           .score.mean().reset_index().groupby(
            'gender').score.mean() * 100
        res = Munch(
            total=student_score, grade=grade, code=code, category=category,
            **gender_score.to_dict())
        return res

    @staticmethod
    def _count_school_score(group, key: str):
        """计算学校的维度、领域性别分数"""
        dim_score = group.loc[group.category == 'dimension', :][key].mean()
        field_score = group.loc[group.category == 'field', :][key].mean()
        return (dim_score + field_score) / 2

    def _count_cls(self, row):
        if row['cls']:
            return f"{project.grade_simple[row['grade']]}({int(row['cls'])})班"
        else:
            return f"{project.grade_simple[row['grade']]}年级"

    def _retain_prec(self, n: float, prec: int = 1):
        n = n * 100
        return int(n) if n in (0, 0.0, 100.0, 100) else round(n, prec)

    def _get_value(self, v, k):
        """计算去年数据中的总人数"""
        base: Union[dict, int] = dict(v).get('people', 0)
        if base:
            return base[k]
        else:
            return 0

    @staticmethod
    def _count_year_score(category: str, data: dict):
        """
        :param category: total/M/F
        :param data: 需计算数据
        """
        dim_score = sum([dict(v)[category] for v in list(data.values()) if v['category'] == 'dimension']) / 4
        field_score = sum([dict(v)[category] for v in list(data.values()) if v['category'] == 'field']) / 6
        score = (dim_score + field_score) / 2
        return score



from collections import namedtuple

import pandas as pd
from loguru import logger

from hawa.base.db import DbUtil
from hawa.base.decos import singleton


@singleton
class DataQuery:
    db = DbUtil()

    def query_unit(self, meta_unit_type: str, meta_unit_id: str):
        MetaUnit = namedtuple("MetaUnit", ["id", "name", "short_name"])
        match meta_unit_type:
            case 'school':
                sql = f"select id,name,short_name from schools where id={meta_unit_id};"
                self.db.cursor.execute(sql)
                data = self.db.cursor.fetchone()
                meta_unit = MetaUnit(**data)
            case 'district' | 'city' | 'province':
                sql = f"select id,name from locations where id={meta_unit_id};"
                self.db.cursor.execute(sql)
                data = self.db.cursor.fetchone()
                meta_unit = MetaUnit(**data, short_name=data['name'])
            case _:
                meta_unit = MetaUnit(id=0, name='全国', short_name='全国')
        return meta_unit

    def query_schools_all(self):
        sql = f"select * from schools;"
        return pd.read_sql(sql, self.db.conn)

    def query_schools_by_ids(self, school_ids: list[int]):
        if len(school_ids) == 0:
            return []
        elif len(school_ids) == 1:
            sql = f"select * from schools where id={school_ids[0]};"
        else:
            sql = f"select * from schools where id in {tuple(school_ids)};"
        return pd.read_sql(sql, self.db.conn)

    def query_schools_by_startwith(self, startwith: int):
        while startwith % 10 == 0:
            startwith = startwith // 10
            logger.info(f"startwith: {startwith}")

        param_len = len(str(startwith))
        sql = f"select * from schools where left(id,{param_len})={startwith};"
        return pd.read_sql(sql, self.db.conn)

    def query_papers(self, test_type: str = '', test_types: list[str] = None):
        """优先 test_types"""
        if test_types:
            sql = f"select * from papers where test_type in {tuple(test_types)};"
        elif test_type:
            sql = f"select * from papers where test_type='{test_type}';"
        else:
            raise
        return pd.read_sql(sql, self.db.conn)

    def query_cases(
            self, school_ids: list[int], paper_ids: list[int],
            valid_to_start: str, valid_to_end: str,
    ):
        if len(paper_ids) == 0:
            return []
        elif len(paper_ids) == 1:
            paper_sql = f"and c.paper_id={paper_ids[0]}"
        else:
            paper_sql = f"and c.paper_id in {tuple(paper_ids)}"

        if len(school_ids) == 0:
            return []
        elif len(school_ids) == 1:
            school_sql = f"and cs.school_id={school_ids[0]}"
        else:
            school_sql = f"and cs.school_id in {tuple(school_ids)}"

        sql = f"select c.id,c.name,c.valid_from,c.valid_to,c.client_id,c.created," \
              f"c.paper_id,c.is_cleared " \
              f"from cases c " \
              f"left join case_schools cs on c.id=cs.case_id " \
              f"where valid_to between '{valid_to_start}' and '{valid_to_end}'" \
              f" {school_sql} {paper_sql} and is_cleared=1;"
        cases = pd.read_sql(sql, self.db.conn).drop_duplicates(subset=['id'])
        return cases

    def query_answers(self, case_ids: list[int]):
        if len(case_ids) == 0:
            return []
        elif len(case_ids) == 1:
            sql = f"select * from answers where case_id={case_ids[0]} and valid=1;"
        else:
            sql = f"select * from answers where case_id in {tuple(case_ids)} and valid=1;"
        answers = pd.read_sql(sql, self.db.conn).drop_duplicates(subset=['case_id', 'student_id', 'item_id'])
        return answers

    def query_students(self, student_ids: list[int]):
        sql = f"select * from users where id in {tuple(student_ids)};"
        students = pd.read_sql(sql, self.db.conn).drop_duplicates(subset=['id'])
        return students

    def query_items(self, item_ids: list[int]):
        sql = f"select * from items where id in {tuple(item_ids)}"
        return pd.read_sql(sql, self.db.conn)

    def query_item_codes(self, item_ids: list[int]):
        item_code_sql = f'select ic.item_id,ic.code,ic.category,c.name ' \
                        f'from item_codes ic left join codebook c on ic.code = c.code ' \
                        f'where ic.item_id in {tuple(item_ids)};'
        item_codes = pd.read_sql(item_code_sql, self.db.conn)
        return item_codes

    @property
    def conn(self):
        return self.db.conn



from dataclasses import dataclass
from typing import Optional

import pandas as pd

from hawa.config import project
from hawa.base.db import DbUtil


class Measurement:
    db = DbUtil()

    @property
    def level_names(self):
        return '、'.join(project.ranks['FEEDBACK_LEVEL'].values())

    @property
    def level_columns(self):
        data = project.ranks['FEEDBACK_LEVEL']
        res = [data[i] for i in sorted(data.keys())]
        return res

    @property
    def field_names(self):
        return self._get_field_names('field')

    @property
    def dimension_names(self):
        return self._get_field_names('dimension')

    @property
    def fields(self):
        return self._get_fields(category='field')

    @property
    def dimensions(self):
        return self._get_fields(category='dimension')

    def _get_fields(self, category: str):
        data = pd.read_sql(
            f"select * from codebook where category='{category}' and name<>'其他'", self.db.conn)
        return data['name'].to_list()

    def _get_field_names(self, category: str):
        codes = self._get_fields(category=category)
        return '、'.join(codes[:len(codes) - 1]) + f'与{codes[-1]}'


@dataclass
class CaseData:
    cases: pd.DataFrame = pd.DataFrame()

    @property
    def join_date(self):
        valid_from = self.cases['valid_from'].min()
        valid_to = self.cases['valid_to'].max()
        if (valid_from.month, valid_from.day) == (valid_to.month, valid_to.day):
            join_date = f"参测日期：{valid_from:%Y年%m月%d日}"
        else:
            join_date = f"参测日期：{valid_from:%Y年%m月%d日}~{valid_to:%m月%d日}"
        return join_date

    @property
    def start_date(self) -> str:
        return self.join_date.split('~')[0].removeprefix('参测日期：')

    @property
    def end_date(self):
        if '~' in self.join_date:
            temp = self.join_date.split('~')[1]
        else:
            temp = self.join_date
        return temp.removeprefix('参测日期：')

    @property
    def start_month(self):
        return self.start_date.split('月')[0] + '月'


@dataclass
class GradeData:
    case_ids: list[int]
    grades: Optional[list[int]] = None
    is_single_grade: bool = False

    def __post_init__(self):
        self.grades = sorted(set(int(str(i)[-2:]) for i in self.case_ids))
        self.is_single_grade = len(self.grades) == 1

    @property
    def grade_name(self) -> str:
        return '、'.join([project.grade_simple[i] for i in self.grades]) + '年级'

    @property
    def grade_name_list(self) -> list[str]:
        return [f"{project.grade_simple[i]}年级" for i in self.grades]



from loguru import logger

from hawa.paper.health import HealthReportData
from test.mock import prepare_test

prepare_test()


def test_health_init():
    rows = [
        {"meta_unit_type": "school", "meta_unit_id": 3707030003, "target_year": 2021},
        # {"meta_unit_type": "district", "meta_unit_id": 370703, "target_year": 2021},
        # {"meta_unit_type": "city", "meta_unit_id": 331100, "target_year": 2021},
        # {"meta_unit_type": "province", "meta_unit_id": 410000, "target_year": 2021},
    ]
    for row in rows:
        logger.info(row)
        HealthReportData(**row)



import setuptools

from verion_text import text

with open("README.md", "r", encoding='utf8') as fh:
    long_description = fh.read()
    for version, description in text:
        long_description += f"\n- {version} {description}"

with open("requirements.txt", "r", encoding='utf8') as fh:
    requires = fh.read().replace('==', '>=')

setuptools.setup(
    name="HawaData",
    version=text[-1][0],
    author="Steven Wang",
    author_email="brightstar8284@icloud.com",
    description="Use For Unified Hawa Data.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/StevenLianaL/HawaData",
    packages=setuptools.find_packages(exclude=["tests*"]),
    install_requires=requires,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ],
)



from hawa.base.db import MongoUtil
from hawa.config import project


def prepare_test():
    project.COMPLETED = True
    MongoUtil.connect()


def validate_data_for_web(d):
    """校验数据是否可以被 web 响应"""
    match str(type(d)):
        case "<class 'str'>" | "<class 'int'>" | "<class 'float'>":
            pass
        case "<class 'list'>":
            for item in d:
                validate_data_for_web(item)
        case "<class 'dict'>" | "<class 'collections.defaultdict'>":
            for v in d.values():
                validate_data_for_web(v)
        case _:
            raise TypeError(f'不支持的类型 {type(d)}')



import MySQLdb
import mongoengine
import sqlalchemy
from databases import Database
from redis.client import StrictRedis

from hawa.base.decos import singleton
from hawa.config import project

DATABASE_URL = f"{project.DB_MODE}://{project.DB_USER}:{project.DB_PSWD}@{project.DB_HOST}/" \
               f"{project.DB_NAME}?charset=utf8"

database = Database(DATABASE_URL)

metadata = sqlalchemy.MetaData()

engine = sqlalchemy.create_engine(DATABASE_URL, encoding='utf-8')


@singleton
class DbUtil:
    _conn = None
    _cursor = None

    @property
    def conn(self):
        return engine

    def connect(self):
        return MySQLdb.connect(
            host=project.DB_HOST,
            port=project.DB_PORT,
            user=project.DB_USER,
            passwd=project.DB_PSWD,
            db=project.DB_NAME,
        )

    @property
    def cursor(self):
        if not self._cursor:
            conn = self.connect()
            self._cursor = conn.cursor(cursorclass=MySQLdb.cursors.DictCursor)
        return self._cursor


class MongoUtil:
    @classmethod
    def connect(self):
        mongoengine.connect(
            project.MONGO_DB,
            host=project.MONGO_HOST, port=project.MONGO_PORT,
            username=project.MONGO_USER,
            password=project.MONGO_PSWD,
            authentication_source=project.MONGO_AUTH_DB
        )


@singleton
class RedisUtil:
    @property
    def conn(self):
        return StrictRedis(
            host=project.REDIS_HOST,
            db=project.REDIS_DB,
            decode_responses=True,
        )



from collections import defaultdict
from dataclasses import dataclass, field
from typing import Set

import pandas as pd

from hawa.common.data import CommonData
from hawa.config import project


@dataclass
class MhtData(CommonData):
    test_type: str = 'mht'
    code_word_list: Set[str] = field(default_factory=lambda: {'mht'})

    # 计算数据
    scale_student_score: dict = field(default_factory=dict)
    sub_scale_score: dict = field(default_factory=dict)
    grade_scale_student_score: dict = field(default_factory=list)
    grade_sub_scale_score: dict = field(default_factory=dict)
    grade_special_students: dict = field(default_factory=dict)

    # 计算数据2
    invalid_student_count: int = 0
    unused_student_count: int = 0
    unused_student_ids: list[int] = field(default_factory=list)
    mht_final_answers: pd.DataFrame = pd.DataFrame()
    mht_final_scores: pd.DataFrame = pd.DataFrame()

    def _to_count_a_final_answers(self):
        super()._to_count_a_final_answers()
        old_answers = self.final_answers
        old_answers['score'] = old_answers['score'].astype(int)

        # 去除具备社会期许效应的学生
        to_count_validity_answers = old_answers.loc[old_answers['mht'] == '效度', :]
        unused_student_ids = []
        for student_id, group in to_count_validity_answers.groupby(by='student_id'):
            if group['score'].sum() >= 8:
                unused_student_ids.append(student_id)

        self.unused_student_ids = unused_student_ids
        self.unused_student_count = len(unused_student_ids)
        self.total_student_count = self.student_count + self.unused_student_count

        self.final_answers = old_answers[~old_answers['student_id'].isin(unused_student_ids)]

        project.logger.debug(f'final_answers: {len(self.final_answers)}')

    def _to_count_c_mht_ans_score(self):
        self.mht_final_answers = self.final_answers.loc[self.final_answers['mht'] != '效度', :]
        self.mht_final_scores = self.count_final_score(answers=self.mht_final_answers)
        self.mht_final_scores['score'] = self.mht_final_scores['score'].astype(int)

    def _to_count_d_scale_student_score(self):
        """学生总量表得分图数据，横轴分数，纵轴人数 （总 及 各年级）"""
        self.scale_student_score = self._tool_count_student_score(score=self.mht_final_scores)

    def _to_count_e_sub_scale_code_score(self):
        """在 8 个子量表上的得分图，横轴量表，纵轴分数"""
        self.sub_scale_score = self._tool_count_sub_code_score(
            answers=self.mht_final_answers
        )

    def _to_count_f_grade_student_score(self):
        """参考 _to_count_c_student_score， 分年级计算"""
        res = {}
        for grade, grade_group in self.mht_final_scores.groupby(by='grade'):
            grade_data = self._tool_count_student_score(score=grade_group)
            res[grade] = grade_data
        self.grade_scale_student_score = res

    def _to_count_g_grade_sub_scale_code_score(self):
        """在 8 个子量表上的得分图，横轴量表，纵轴分数"""
        res = {}
        for grade, grade_ans_group in self.mht_final_answers.groupby('grade'):
            grade_data = self._tool_count_sub_code_score(answers=grade_ans_group)
            res[grade] = grade_data
        self.grade_sub_scale_score = res

    def _to_count_h_grade_special_students(self):
        """计算各年级 某量表超过8分的学生"""
        res = defaultdict(list)
        for grade, grade_ans_group in self.mht_final_answers.groupby('grade'):
            for student_id, student_group in grade_ans_group.groupby('student_id'):
                student_name = student_group['username'].tolist()[0]
                res[grade].append(
                    self._tool_count_sub_code_score(answers=student_group, unit_name=student_name)
                )
        self.grade_special_students = res

    # 计算工具
    def _tool_count_student_score(self, score: pd.DataFrame):
        data = []
        handred = set(range(1, 101))
        for score, row in score.groupby('score'):
            handred.discard(score)
            data.append((score, int(row.score.count())))
        for score in handred:
            data.append((score, 0))
        data.sort(key=lambda x: x[0])
        x_axis, y_axis = [], []
        for (score, student_count) in data:
            x_axis.append(score)
            y_axis.append(student_count)

        return {
            "name": self.meta_unit.name,
            "x_axis": x_axis,
            "y_axis": y_axis
        }

    def _tool_count_sub_code_score(self, answers: pd.DataFrame, unit_name: str = ''):
        mht_scores = defaultdict(list)
        x_axis, y_axis = [], []
        for (student_id, mht), group in answers.groupby(by=['student_id', 'mht']):
            if mht == '效度':
                continue
            mht_scores[mht].append(group.score.sum())
        for mht, score_list in mht_scores.items():
            x_axis.append(mht)
            y_axis.append(round(float(sum(score_list) / len(score_list)), 1))

        return {
            "name": unit_name if unit_name else self.meta_unit.name,
            "x_axis": x_axis,
            "y_axis": y_axis,
        }


@dataclass
class MhtWebData(MhtData):
    pass
